import sys
from thread import get_ident, start_new_thread
from threading import currentThread as current_thread, _active, _active_limbo_lock, _limbo, _MainThread, _DummyThread
import threading

from mext.log import Logger
from mext.classhacks import get_name
from mext.context import State

from mext.threads.events import Flag

__all__ = ['Thread']



class Thread(object):
    state = None
    log = Logger('threads')
    ignore_errors = False
    #_exc_clear = sys.exc_clear

    def __init__(self, func, *args, **kw):
        self.func = func
        self.name = get_name(func, 'Thread-%s' % func)
        self.args = args
        self.kw = kw

        self.daemon = current_thread().daemon
        self.id = None
        self.started = Flag()
        self.stopped = Flag()

    def __repr__(self):
        if self.stopped:
            status = "stopped"
        elif self.started:
            status = "started"
        else:
            status = "new"
        if self.daemon:
            status += " daemon"
        if self.id is not None:
            status += " %s" % self.id
        return "<%s(%s, %s)>" % (self.__class__.__name__, self.name, status)


    is_main = False
    get = get_current = staticmethod(current_thread)

    @classmethod
    def find(self, id):
        return _active.get(id)



    @property
    def is_current(self):
        return self.id == get_ident()

    def start(self):
        if self.state is None:
            self.state = State.clone()
        #self.log.info("Starting new thread %s", self.name)
        if self.started:
            raise RuntimeError("thread already started")
        with _active_limbo_lock:
            _limbo[self] = self
        try:
            start_new_thread(self._thread_main, ())
        except:
            with _active_limbo_lock:
                del _limbo[self]
            raise
        self.started.wait()
        return self

    def start_daemon(self):
        self.daemon = True
        return self.start()



    def _thread_main(self):
        try:
            try:
                sys.settrace(threading._trace_hook)
                self.started.set()
                self.id = get_ident()
                with _active_limbo_lock:
                    _active[self.id] = self
                    del _limbo[self]
                self._thread_main_inner()
            finally:
                with _active_limbo_lock:
                    _active.pop(get_ident(), None)
                self.stopped.set()
        except:
            if not self.daemon or sys is not None:
                raise


    def _thread_main_inner(self):
        with self.state:
            try:
                #self.log.info("Thread %s is running", self.name)
                self.func(*self.args, **self.kw)
            except SystemExit:
                raise
            except:
                self._handle_error()
            finally:
                #self.log.debug("Thread stopped")
                # Avoid a refcycle if the thread is running a function with
                # an argument that has a member that points to the thread.
                del self.func, self.args, self.kw
                #self._exc_clear()

    def _handle_error(self):
        if sys is None:
            raise
        exc_info = sys.exc_info()
        if not self.ignore_errors:
            sys.excepthook(*exc_info)
        #elif isinstance(self.ignore_errors, Logger)::
        elif self.ignore_errors == 'log':
            self.log.error(u"Thread raised an exception", exc_info=True)
        else:
            self.log.debug(u"Thread raised an exception %s", exc_info[1])

    def join(self, timeout=None):
        if not self.started:
            raise RuntimeError("cannot join thread before it is started")
        if self.is_current:
            raise RuntimeError("cannot join current thread")
        self.stopped.wait(timeout)



    # this makes threading._after_fork work
    def __stopped(self, v):
        self.stopped.set()
    __stopped = property(fset=__stopped)

    # required for threading._pickSomeNonDaemonThread
    def is_alive(self):
        return self.started and not self.stopped

    alive = property(is_alive)

    # fix for mext.log
    def getName(self):
        return self.name


_MainThread.is_main = True
_MainThread.is_current = Thread.is_current
_MainThread.id = get_ident()
_DummyThread.is_main = False

